//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//
// "Creature Box" -- flocking app
//
// by Christopher Rasmussen
//
// CISC 440/640, March, 2014
//
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include "Flocker.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

int flocker_history_length = 10;
int flocker_draw_mode = DRAW_MODE_POLY;
vector <Flocker *> flocker_array;    
vector <vector <double> > flocker_squared_distance;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// attempt to be slightly efficient by pre-calculating all of the distances between
// pairs of flockers exactly once

void calculate_flocker_squared_distances()
{
  int i, j;
  Vec3 diff;

  for (i = 0; i < flocker_array.size(); i++)
    for (j = i + 1; j < flocker_array.size(); j++) {
      diff.subtract(flocker_array[i]->position, flocker_array[j]->position);
      flocker_squared_distance[i][j] = diff.squaredNorm();
    }
}

//----------------------------------------------------------------------------

Flocker::Flocker(int _index,
		 double init_x, double init_y, double init_z,
		 double init_vx, double init_vy, double init_vz,
		 double rand_force_limit,
		 double min_separate_distance, double max_separate_distance,  double separate_weight,
		 double min_align_distance, double max_align_distance, double align_weight,
		 double min_cohere_distance, double max_cohere_distance, double cohere_weight,
		 float r, float g, float b,
		 int max_hist) : Creature(_index, init_x, init_y, init_z, init_vx, init_vy, init_vz, r, g, b, max_hist)
{ 
  random_force_limit = rand_force_limit;

  separation_weight = separate_weight;

  min_separation_distance = min_separate_distance;
  min_squared_separation_distance = min_separation_distance * min_separation_distance;

  max_separation_distance = max_separate_distance;
  max_squared_separation_distance = max_separation_distance * max_separation_distance;

  inv_range_squared_separation_distance = 1.0 / (max_squared_separation_distance - min_squared_separation_distance);

  alignment_weight = align_weight;

  min_alignment_distance = min_align_distance;
  min_squared_alignment_distance = min_alignment_distance * min_alignment_distance;

  max_alignment_distance = max_align_distance;
  max_squared_alignment_distance = max_alignment_distance * max_alignment_distance;

  inv_range_squared_alignment_distance = 1.0 / (max_squared_alignment_distance - min_squared_alignment_distance);

  cohesion_weight = cohere_weight;

  min_cohesion_distance = min_cohere_distance;
  min_squared_cohesion_distance = min_cohesion_distance * min_cohesion_distance;

  max_cohesion_distance = max_cohere_distance;
  max_squared_cohesion_distance = max_cohesion_distance * max_cohesion_distance;

  inv_range_squared_cohesion_distance = 1.0 / (max_squared_cohesion_distance - min_squared_cohesion_distance);
}

//----------------------------------------------------------------------------

void Flocker::draw()
{
  if (flocker_draw_mode == DRAW_MODE_HISTORY) {

    // position "trail" using history

    glPointSize(3);

    float inv_size = 1.0 / position_history.size();
    
    glBegin(GL_POINTS);
    
    float index = position_history.size();
    for (deque<Vec3>::iterator it = position_history.begin(); it!=position_history.end(); ++it) {
      glColor3f(draw_color.r * index * inv_size, draw_color.g * index * inv_size, draw_color.b * index * inv_size);
      glVertex3f((*it).x, (*it).y, (*it).z);
      index--;
    }
    
    glEnd();
  }

  // local coordinate frame axes

  else if (flocker_draw_mode == DRAW_MODE_AXES) {

    double axis_scale = 0.25;

    glBegin(GL_LINES);
    
    glColor3f(1, 0, 0);
    glVertex3f(position.x, position.y, position.z);
    glVertex3f(position.x + axis_scale * frame_x.x, position.y + axis_scale * frame_x.y, position.z + axis_scale * frame_x.z);
    
    glColor3f(0, 1, 0);
    glVertex3f(position.x, position.y, position.z);
    glVertex3f(position.x + axis_scale * frame_y.x, position.y + axis_scale * frame_y.y, position.z + axis_scale * frame_y.z);
    
    glColor3f(0, 0, 1);
    glVertex3f(position.x, position.y, position.z);
    glVertex3f(position.x + axis_scale * frame_z.x, position.y + axis_scale * frame_z.y, position.z + axis_scale * frame_z.z);
    
    glEnd();
  }

  // flocker 3-D shape

  else if (flocker_draw_mode == DRAW_MODE_POLY) {

    double width = 0.2;
    double height = 0.2;
    double length = 0.3;
  
    // horizontal
    
    glColor3f(1, 0, 0);
    
    glBegin(GL_POLYGON);
    
    glVertex3f(position.x - 0.5 * width * frame_x.x, 
	       position.y - 0.5 * width * frame_x.y, 
	       position.z - 0.5 * width * frame_x.z);
    
    glVertex3f(position.x - 0.8 * length * frame_z.x, 
	       position.y - 0.8 * length * frame_z.y, 
	       position.z - 0.8 * length * frame_z.z);
    
    glVertex3f(position.x + 0.5 * width * frame_x.x, 
	       position.y + 0.5 * width * frame_x.y, 
	       position.z + 0.5 * width * frame_x.z);
    
    glVertex3f(position.x + 0.2 * length * frame_z.x, 
	       position.y + 0.2 * length * frame_z.y, 
	       position.z + 0.2 * length * frame_z.z);
    
    glEnd();
    
    // vertical
    
    glColor3f(0, 1, 0);
    
    glBegin(GL_POLYGON);
    
    glVertex3f(position.x + 0.5 * height * frame_y.x, 
	       position.y + 0.5 * height * frame_y.y, 
	       position.z + 0.5 * height * frame_y.z);
    
    glVertex3f(position.x - 0.8 * length * frame_z.x, 
	       position.y - 0.8 * length * frame_z.y, 
	       position.z - 0.8 * length * frame_z.z);
    
    glVertex3f(position.x - 0.5 * height * frame_y.x, 
	       position.y - 0.5 * height * frame_y.y, 
	       position.z - 0.5 * height * frame_y.z);
    
    glVertex3f(position.x + 0.2 * length * frame_z.x, 
	       position.y + 0.2 * length * frame_z.y, 
	       position.z + 0.2 * length * frame_z.z);
    
    glEnd();
  }
}

//----------------------------------------------------------------------------

// based on:
// http://processing.org/examples/flocking
// http://libcinder.org/docs/dev/flocking_chapter2.html

// side effect is putting values into SEPARATION_FORCE vector

bool Flocker::compute_separation_force()
{
  int j;
  Vec3 direction;
  int count = 0;
  double mag;
  double F;

  separation_force = Vec3(0, 0, 0);

  for (j = index + 1; j < flocker_squared_distance.size(); j++)
    if (flocker_squared_distance[index][j] >= min_squared_separation_distance &&
	flocker_squared_distance[index][j] <= max_squared_separation_distance) {

      // set (unweighted) force magnitude

      F = max_squared_separation_distance / flocker_squared_distance[index][j] - 1.0;

      // set force direction

      direction.subtract(position, flocker_array[j]->position);
      direction.normalize();
      direction.multiply(F);
      separation_force.add(direction);
      count++;
    }
  for (j = index - 1; j >= 0; j--)
    if (flocker_squared_distance[j][index] >= min_squared_separation_distance && 
	flocker_squared_distance[j][index] <= max_squared_separation_distance) {

      // set (unweighted) force magnitude

      F = max_squared_separation_distance / flocker_squared_distance[j][index] - 1.0;

      // set force direction

      direction.subtract(position, flocker_array[j]->position);
      direction.normalize();
      direction.multiply(F);
      separation_force.add(direction);
      count++;
    }

  if (count > 0) {
    separation_force.multiply(separation_weight);
    return true;
  }
  else
    return false;
}

//----------------------------------------------------------------------------

// based on:
// http://processing.org/examples/flocking
// http://libcinder.org/docs/dev/flocking_chapter4.html

// side effect is putting values into ALIGNMENT_FORCE vector

bool Flocker::compute_alignment_force()
{
  int j;
  Vec3 direction;
  int count = 0;
  double mag, percent;
  double F;

  alignment_force = Vec3(0, 0, 0);

  for (j = index + 1; j < flocker_squared_distance.size(); j++)
    if (flocker_squared_distance[index][j] >= min_squared_alignment_distance &&
	flocker_squared_distance[index][j] <= max_squared_alignment_distance) {

      // set (unweighted) force magnitude

      percent = (flocker_squared_distance[index][j] - max_squared_alignment_distance) * inv_range_squared_alignment_distance;
      F = 0.5 + -0.5 * cos(percent * 2.0 * M_PI);

      // set force direction

      direction = flocker_array[j]->velocity;
      direction.normalize();
      direction.multiply(F);
      alignment_force.add(direction);   
      count++;
    }
  for (j = index - 1; j >= 0; j--)
    if (flocker_squared_distance[j][index] >= min_squared_alignment_distance &&
	flocker_squared_distance[j][index] <= max_squared_alignment_distance) {

      // set (unweighted) force magnitude

      percent = (flocker_squared_distance[index][j] - max_squared_alignment_distance) * inv_range_squared_alignment_distance;
      F = 0.5 + -0.5 * cos(percent * 2.0 * M_PI);

      // set force direction

      direction = flocker_array[j]->velocity;
      direction.normalize();
      direction.multiply(F);
      alignment_force.add(direction);   
      count++;
    }

  if (count > 0) {
    alignment_force.multiply(alignment_weight);
    return true;
  }
  else
    return false;
}

//----------------------------------------------------------------------------

// based on:
// http://processing.org/examples/flocking 
// http://libcinder.org/docs/dev/flocking_chapter3.html

// side effect is putting values into COHESION_FORCE vector

bool Flocker::compute_cohesion_force()
{
  int j;
  Vec3 direction;
  int count = 0;
  double mag, percent;
  double F;

  cohesion_force = Vec3(0, 0, 0);

  for (j = index + 1; j < flocker_squared_distance.size(); j++)
    if (flocker_squared_distance[index][j] >= min_squared_cohesion_distance &&
	flocker_squared_distance[index][j] <= max_squared_cohesion_distance) {

      // set (unweighted) force magnitude

      percent = (flocker_squared_distance[index][j] - max_squared_cohesion_distance) * inv_range_squared_cohesion_distance;
      F = 0.5 + -0.5 * cos(percent * 2.0 * M_PI);

      // set force direction

      direction.subtract(flocker_array[j]->position, position);    // opposite direction of separation
      direction.normalize();
      direction.multiply(F);
      cohesion_force.add(direction);   
      count++;
    }
  for (j = index - 1; j >= 0; j--)
    if (flocker_squared_distance[j][index] >= min_squared_cohesion_distance &&
	flocker_squared_distance[j][index] <= max_squared_cohesion_distance) {

      // set (unweighted) force magnitude

      percent = (flocker_squared_distance[index][j] - max_squared_cohesion_distance) * inv_range_squared_cohesion_distance;
      F = 0.5 + -0.5 * cos(percent * 2.0 * M_PI);

      // set force direction

      direction.subtract(flocker_array[j]->position, position);   // opposite direction of separation
      direction.normalize();
      direction.multiply(F);
      cohesion_force.add(direction);   
      count++;
    }

  if (count > 0) {
    cohesion_force.multiply(cohesion_weight);
    return true;
  }
  else
    return false;
}

//----------------------------------------------------------------------------

// apply physics

void Flocker::update()
{
  // set accelerations (aka forces)

  acceleration = Vec3(0, 0, 0);
  
  // deterministic behaviors

  compute_separation_force();
  acceleration.add(separation_force);

  compute_alignment_force();
  acceleration.add(alignment_force);

  compute_cohesion_force();
  acceleration.add(cohesion_force);

  draw_color.r = separation_force.norm();
  draw_color.g = alignment_force.norm();
  draw_color.b = cohesion_force.norm();
  if (draw_color.r > 0 || draw_color.g > 0 || draw_color.b > 0)
    draw_color.normalize();
  else 
    draw_color = base_color;

  // randomness

  if (random_force_limit > 0.0) {
    acceleration.x += uniform_random(-random_force_limit, random_force_limit);
    acceleration.y += uniform_random(-random_force_limit, random_force_limit);
    acceleration.z += uniform_random(-random_force_limit, random_force_limit);
  }

  // update velocity

  new_velocity.add(velocity, acceleration);

  // limit velocity

  double mag = new_velocity.magnitude();
  if (mag > MAX_FLOCKER_SPEED)
    new_velocity.divide(mag / MAX_FLOCKER_SPEED);

  // update position

  new_position.add(position, new_velocity);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
