#ifndef UTILS_HH

#define UTILS_HH

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//
// "Creature Box" -- flocking app
//
// by Christopher Rasmussen
//
// CISC 440/640, March, 2014
//
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <vector>
#include <deque>
#include <string>

#include <sys/time.h>
#include <unistd.h>

#include <GL/glut.h>

using namespace std;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void initialize_random();
double uniform_random(double, double);
int uniform_random_int(int, int);
void draw_string(float, float, float, char *);
double get_timestamp();

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#define MAX2(A, B)          (((A) > (B)) ? (A) : (B))

#ifndef DEG2RAD
#define DEG2RAD(x)            ((x) * 0.017453)
#endif

#ifndef RAD2DEG
#define RAD2DEG(x)            ((x) * 57.29578)
#endif


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

class Vec3
{
public:

  union {
    double x;
    double r;
  };

  union {
    double y;
    double g;
  };

  union {
    double z;
    double b;
  };

  Vec3() { }
  Vec3(double _x, double _y, double _z) { x = _x; y = _y; z = _z; }

  void multiply(double val) { x *= val; y *= val; z *= val; }
  void divide(double val) { x /= val; y /= val; z /= val; }

  void add(Vec3 v) { x += v.x; y += v.y; z += v.z; }
  void add(Vec3 v1, Vec3 v2) { x = v1.x + v2.x; y = v1.y + v2.y; z = v1.z + v2.z; }

  void subtract(Vec3 v) { x -= v.x; y -= v.y; z -= v.z; }
  void subtract(Vec3 v1, Vec3 v2) { x = v1.x - v2.x; y = v1.y - v2.y; z = v1.z - v2.z; }

  void cross(Vec3 v1, Vec3 v2) {
    x = v1.y*v2.z - v1.z*v2.y;
    y = v1.z*v2.x - v1.x*v2.z;
    z = v1.x*v2.y - v1.y*v2.x; 
  }

  double dot(Vec3 v) { return x*v.x + y*v.y + z*v.z; }
  void normalize() { double mag = norm(); x /= mag; y /= mag; z /= mag; }

  double squaredNorm() { return x*x + y*y + z*z; }
  double norm() { return sqrt(squaredNorm());  }
  double magnitude() { return norm(); }

  void print(char * s) { printf("%s: %.3lf %.3lf %.3lf\n", s, x, y, z); }
};

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#endif
